/*
 * Copyright (c) 2021 PIXILAB Technologies AB, Sweden (http://pixilab.se). All Rights Reserved.
 */

import {field, id, list} from "system_lib/Metadata";
import * as feed from "system_lib/Feed";
import {ListData} from "system_lib/Feed";
import {SimpleHTTP} from "system/SimpleHTTP";


//  API GraphQL query endpoint URL for requesting data fmr database
const kQueryUrl = "https://api.maas.museum/graphql?query=";

/** Main class establishing the feed instances s provided by me. I just pick a number
	of interesting collections, giving them a short name that can be referred to
	from within Blocks.
*/
export class Maas extends feed.Feed {
	constructor(env: feed.FeedEnv) {
		super(env);
		this.establishFeed(new Collection("Matchbox", "Matchbox Toys")); // ***
		this.establishFeed(new Collection("Clyde", "Clyde Engineering Photograph"));
		this.establishFeed(new Collection("Observatory", "Sydney Observatory"));
		this.establishFeed(new Collection("Wunderlich", "Wunderlich Limited Archive"));
		this.establishFeed(new Collection("Stewart", "A A Stewart Engineering Model")); // ***
		this.establishFeed(new Collection("Lennon", "Tom Lennon Photographic Archive"));
	}
}

/**
 * Manage a specific "collection" as its own feed instance, making it easy for
 * different display spots to show different collections. I just pick the first
 * objects in each collection as they exist in the database. In a real-world
 * use case, the database should contain some flag or priority value or other
 * criteria that can be selected for and sorted on for the desired set of
 * objects to appear on screen.
 */
class Collection implements feed.StaticFeed<ListItem, DetailsItem> {
	readonly name: string;				// Our internal, mandatory and short feed name
	readonly collectionName: string;	// Official name of feed used by API

	// Mandatory fields describing what types of list and item data I provide
	readonly listType = ListItem;
	readonly itemType = DetailsItem;

	constructor(feedName: string, collectionName: string) {
		this.name = feedName;
		this.collectionName = collectionName;
	}

	/**
	 * Mandatory method used to get a list of items.
	 */
	getList(spec:feed.FeedListSpec): Promise<ListData<ListItem>> {
		/*	Define the GraphQL query to send to get a reasonably sized list of items.
			The request is designed to return a subset of the fields in each item,
			only what's needed to display the list, including a thumbnail version
			of its first image.
		*/
		const graphQl = `{
			objects(limit: 30, skip: 0, filter: {
				collectionName:  "${this.collectionName}",
				hasMedia:true
			}) {
				_id
				displayTitle
				images(limit: 1) {
					thumb: url(width: 250, height: 250, smart:true)
				}
			}
		}`;

		// Use SimpleHTTP to fetch the list from the server, passing it to processList
		return SimpleHTTP.newRequest(
			kQueryUrl + encodeURI(graphQl),
			{ interpretResponse: true }
		)
		.setTimeout(25)	// Yes, it occasionally takes quite a while to respond
		.get<GraphQlListResponse<GraphQlListItem>>()
		.then(response => Collection.processList(response.interpreted.data.objects));
	}

	/**
	 * Convert list of items returned from GraphQL API to our desired ListData form.
	 */
	static processList(list: GraphQlListItem[]): ListData<ListItem> {
		const listItems: ListItem[] = [];
		for (var item of list) {
			// Pass explicit null if has no thumbnail to not request through details
			listItems.push(new ListItem(
				item._id,
				item.displayTitle,
				item.statement,
				Collection.getThumbUrl(item)
			));
		}
		return { items: listItems };
	}

	/** Mandatory method used to get a single item with full details, based
	 	on the ID requested according to spec.
	 */
	getDetails(spec: feed.FeedDetailsSpec) {
		const graphQl = `{
			object(filter: {
				_id: ${spec.id}
			}) {
				_id
				displayTitle
				description
				statement
				images(limit: 6) {
				  large: url(width: 1920)
				  thumb: url(width: 250, height: 250, smart:true)
				}
			}
		}`;
		return SimpleHTTP.newRequest(
			kQueryUrl + encodeURI(graphQl),
			{ interpretResponse: true }
		)
		.setTimeout(25)	// Yes, it occasionally takes quite a while to respond
		.get<GraphQlSingleResponse<GraphQlSingleItem>>()
		.then(response => Collection.processDetails(response.interpreted.data.object));
	}

	/**
	 * Return the first thumbnail (if any) in the images array in item.
	 * Return explicit null if has no thumbnail to differentiate from
	 * undefined which is in places interpreted as "not available"
	 */
	private static getThumbUrl(item: { images: {thumb: string}[] }): string|null {
		const firstImage = item.images[0];
		return firstImage ? firstImage.thumb : null;
	}

	/**
	 * Convert single item returned from GraphQL API to our desired DetailsItem.
	 */
	private static processDetails(item: GraphQlSingleItem): DetailsItem {
		const images: Image[] = [];
		for (var image of item.images)
			images.push(new Image(image.large, image.thumb));
		return new DetailsItem(
			item._id,
			item.displayTitle,
			item.statement,
			Collection.getThumbUrl(item),
			item.description,
			images
		);
	}
}

/**	Shape of each item in the brief list of items returned for
	display in the grid. The amount of data provided here is
	deliberately kept to a bare minimum, since we may need to deal
	with many objects of this type.
*/
class ListItem {

	/*	The @id field type is special in that it defines a unique ID of this item
		that can subsequently be used to fetch the corresponding details for
		that item.
	*/
	@id("Uniquely identifying this feed item")
	readonly id: number;

	/*	Use the @field annotation to make fields visible in Blocks, also providing
		a description that appears in the Blocks editor for this field.
	*/
	@field("Title of the item")
	readonly title: string;

	@field("Small (250x250) thumbnail image URL")
	readonly thumb: string;

	constructor(id: number, title: string, brief: string, thumb: string) {
		this.id = id;
		this.title = title;
		this.thumb = thumb;
	}
}


/**	Shape of a single image, used in DetailsItem since it can provide a
	list of images associated with that single item
*/
class Image {
	constructor(large: string, thumb: string) {
		this.large = large;
		this.thumb = thumb;
	}

	@field("URL to large (1080) image")
	readonly large: string;

	@field("Small (250x250) thumbnail image URL")
	readonly thumb: string;
}

/**	A full details item, fetched to view all desired details about a single
	item, possibly including lists of nested sub-items, such as the Image field.
	Note that the DetailsItem must extend the ListItem, and hence must provide
	that same data, plus its own extended data.
*/
class DetailsItem  extends ListItem {
	@field("Detailed description")
	readonly details: string;

	@field("Brief description")
	readonly brief: string;

	// The @list annotation defines a field that holds a list of data items
	@list(Image, "All images associated with this item")
	readonly images?: Image[];

	constructor(
		id: number, title: string, brief: string, thumb: string,
		details: string,
		images?: Image[],
	) {
		super(id, title, brief, thumb),
		this.brief = brief;
		this.details = details;
		this.images = images;
	}
}


// What's returned from a graphql list query
interface GraphQlListResponse<T> {
	data: {
		objects: T[];	// Actual list of interesting objects
	}
}

// What's returned from a graphql single item query
interface GraphQlSingleResponse<T> {
	data: {
		object: T;	// Actual single, interesting object
	}
}


// Elements in GraphQlListResponse, according to query in getList
interface GraphQlListItem {
	_id: number;
 	displayTitle: string;
 	statement: string;		// Brief description
 	images: [				// Empty or single (due to limit) image
 		{
 			thumb: string;	// Thumbnail URL
 		}
 	]
}

// GraphQlSingleResponse data corresponding to query in getDetails
interface GraphQlSingleItem {
	_id: number;
 	displayTitle: string;
 	statement: string;
	brief: string;			// Brief description
	description: string; 	// Detailed description
 	images: [				// Zeror or more (up to limit) images
 		{
 			thumb: string;	// Thumbnail URL
 			large: string;	// Large image URL
 		}
 	]
}
