/*	A Blocks feed script providing URLs to images in a set of designated directories on the server.
	Can be used to build slideshows dynamically, outside of Blocks, by simply populating those
	directories with suitable images.

 	Copyright (c) 2021 PIXILAB Technologies AB, Sweden (http://pixilab.se). All Rights Reserved.
 */

import {field} from 'system_lib/Metadata';
import { SimpleFile } from 'system/SimpleFile';
import {Feed, FeedEnv, FeedListSpec, StaticFeed} from 'system_lib/Feed';

export class Slideshows extends Feed {
	static readonly pathToSlideshows = '/public/feed/slideshows';

	constructor(env: FeedEnv) {
		super(env);

		// Obtain a list of subdirectories to pathToSlideshows, making a SlideshowFeed for each
		SimpleFile.list(Slideshows.pathToSlideshows, true).then(dirInfo => {
			if (!dirInfo.directories.length)
				console.warn("No slideshow directories found in", Slideshows.pathToSlideshows);
			for (const dirName of dirInfo.directories) {
				this.establishFeed(new SlideshowFeed(
					dirName,
					Slideshows.pathToSlideshows + '/' + dirName
				));
			}
		}).catch(error => console.error("Required directory not found", Slideshows.pathToSlideshows, error));
	}
}

/**
 * A single "feed", named by the directory containing its slides. Thus, to add more
 * slideshows, just add one more directory inside pathToSlideshows, with its images.
 */
class SlideshowFeed implements StaticFeed<SlideImageData, SlideImageData> {
	readonly listType = SlideImageData;
	readonly itemType = SlideImageData;

	constructor(
		readonly name: string,
		private readonly pathToSlides: string
	) {
		console.log("Established feed", name);
	}

	async getList(spec: FeedListSpec) {
		const filePaths = await SimpleFile.list(this.pathToSlides);
		/*	Use makeJSArray to convert the "array like" list of files to a true
			JS array in order to apply JS array functions such as sort and map.
		 */
		const sortedPaths = Feed
			.makeJSArray(filePaths.files)
			.sort()
			.map(path => new SlideImageData(path));
		return { items: sortedPaths };
	}

	/*	Since listType and itemType are both the same, there's no need
		to implement any getDetails function here.
	 */
}

/**
 * Data used by each individual slide. Just provides the URL from where to load the image.
 * This is used both as "listType" and "itemType" above, since it's so simple and there's
 * no additional data tpo fetch details for.
 */
class SlideImageData {
	@field("Image URL")
	readonly url: string;

	constructor(url: string) {
		this.url = url;
	}
}
